<?php

namespace jrosset\CommandLine\Argument\Option;

use jrosset\CommandLine\Argument\AbstractArgument;

/**
 * Abstract base for option arguments: --XXX or -X
 */
abstract class AbstractOptionArgument extends AbstractArgument implements IOptionArgument {
    /**
     * @var string|null The short tag (the -X form). Null if none
     */
    private ?string $tagShort;
    /**
     * @var string The long tag (the --XXX form)
     */
    private string $tagLong;

    /**
     * Create a new option argument
     *
     * By default, the {@see AbstractArgument::$varName variable name} is the same as the {@see AbstractArgument::$name argument in help}
     *
     * @param string      $name        The argument name in help
     * @param string|null $description The argument description in help
     * @param string|null $tagLong     The long tag. If null, deduce from argument name <b>$name</b> (see {@see AbstractOptionArgument::convertNameToTagLong()}
     * @param string|null $tagShort    the short tag
     */
    public function __construct (string $name, ?string $description, ?string $tagLong = null, ?string $tagShort = null) {
        parent::__construct($name, $description);

        $this->setTagLong($tagLong ?? static::convertNameToTagLong($name));
        $this->setTagShort($tagShort);
    }

    /**
     * Convert a variable name to a long tag
     *
     * Performs the following tasks :
     *  1. The underscores are replaced by a caret
     *  2. The upper case letters are replaced by a caret followed with the lower case letter
     *
     * @param string $name The name to convert
     *
     * @return string The corresponding long tag
     */
    public static function convertNameToTagLong (string $name): string {
        return preg_replace_callback(
            '/[A-Z_]/',
            function ($matches) {
                return '-' . ($matches[0] === '_' ? '' : '' . mb_strtolower($matches[0]));
            },
            $name
        );
    }

    /**
     * @inheritDoc
     */
    public function getTagShort (): ?string {
        return $this->tagShort;
    }
    /**
     * Set the short tag. Null if none
     *
     * @param string|null $tagShort The short tag
     *
     * @return $this
     */
    public function setTagShort (string $tagShort = null): self {
        $this->tagShort = $tagShort;
        return $this;
    }
    /**
     * Is a short tag defined ?
     *
     * @return boolean True if a short tag is defined, else False
     */
    public function hasTagShort (): bool {
        return $this->getTagShort() !== null;
    }

    /**
     * @inheritDoc
     */
    public function getTagLong (): string {
        return $this->tagLong;
    }
    /**
     * Set the long tag
     *
     * @param string $tagLong The long tag
     *
     * @return $this
     */
    public function setTagLong (string $tagLong): self {
        $this->tagLong = $tagLong;
        return $this;
    }
}