<?php

namespace jrosset\CommandLine\Argument\Option;

use InvalidArgumentException;
use jrosset\CommandLine\Argument\IArgumentValueDescription;
use jrosset\CommandLine\Argument\Parser\IParser;
use UnexpectedValueException;

/**
 * Option argument with possible values
 */
class Value extends AbstractOptionArgument implements IArgumentValueDescription {
    /**
     * @var IParser The values parser
     */
    private IParser $valuesParser;

    /**
     * @var int The minimum number of values
     */
    private int $valueOccurMin;
    /**
     * @var int|null The maximum number of values. Null if unlimited
     */
    private ?int $valueOccurMax;

    /**
     * Create a new option argument
     *
     * By default, the {@see AbstractArgument::$varName variable name} is the same as the {@see AbstractArgument::$name argument in help}
     *
     * @param string      $name         The argument name in help
     * @param string|null $description  The argument name in documentation
     * @param IParser     $valuesParser The value parser
     * @param string|null $tagLong      The long tag. If null, deduce from argument name <b>$name</b> (see {@see AbstractOptionArgument::convertNameToTagLong()}
     * @param string|null $tagShort     The short tag
     */
    public function __construct (string $name, ?string $description, IParser $valuesParser, ?string $tagLong = null, ?string $tagShort = null) {
        parent::__construct($name, $description, $tagLong, $tagShort);

        $this->setValuesParser($valuesParser);
        $this->setValueOccurMin(1);
        $this->setValueOccurMax(1);
    }

    /**
     * @inheritDoc
     */
    public function getValueDescription (): string {
        return $this->valuesParser->getValueDescription();
    }

    /**
     * The values parser
     *
     * @return IParser The values parser
     */
    public function getValuesParser (): IParser {
        return $this->valuesParser;
    }
    /**
     * Set the values parser
     *
     * @param IParser $valuesParser The values parser
     *
     * @return $this
     */
    public function setValuesParser (IParser $valuesParser): self {
        $this->valuesParser = $valuesParser;
        return $this;
    }

    /**
     * Is the value optional ?
     *
     * @return bool Is the value optional ?
     */
    public function isValueOptional (): bool {
        return $this->getValueOccurMin() === 0;
    }
    /**
     * Set if the value is optional
     *
     * @param bool $optional Is the value optional ?
     *
     * @return $this
     */
    public function setValueOptional (bool $optional): self {
        $this->setValueOccurMin($optional ? 0 : 1);
        return $this;
    }

    /**
     * The minimum number of values
     *
     * @return int The minimum number of values
     */
    public function getValueOccurMin (): int {
        return $this->valueOccurMin;
    }
    /**
     * Set the minimum number of values
     *
     * @param int $valueOccurMin The minimum number of values
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the minimum number of values provided is not valid (strictly positive integer)
     * @throws UnexpectedValueException If the minimum number of values provided is strictly greater than the maximum
     */
    public function setValueOccurMin (int $valueOccurMin): self {
        if (!is_numeric($valueOccurMin) || floor($valueOccurMin) != $valueOccurMin || ($valueOccurMin = (int)$valueOccurMin) < 0) {
            throw new InvalidArgumentException('The minimum number of values must be a strictly positive integer');
        }

        $valueOccurMax = $this->getValueOccurMax();
        if ($valueOccurMax !== null && $valueOccurMin > $valueOccurMax) {
            throw new UnexpectedValueException('The minimum number of values provided is strictly greater than the maximum');
        }

        $this->valueOccurMin = $valueOccurMin;
        return $this;
    }

    /**
     * The maximum number of values. Null if unlimited
     *
     * @return int|null The maximum number of value
     */
    public function getValueOccurMax (): ?int {
        return $this->valueOccurMax;
    }
    /**
     * Set the maximum number of values. Null if unlimited
     *
     * @param int|null $valueOccurMax The maximum number of values
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the maximum number of values provided is not valid (strictly positive integer)
     * @throws UnexpectedValueException If the maximum number of values provided is strictly lower than the minimum
     */
    public function setValueOccurMax (?int $valueOccurMax): self {
        if ($valueOccurMax !== null) {
            if (!is_numeric($valueOccurMax) || floor($valueOccurMax) != $valueOccurMax || ($valueOccurMax = (int)$valueOccurMax) < 0) {
                throw new InvalidArgumentException('The maximum number of values must be a strictly positive integer');
            }

            if ($valueOccurMax < $this->getValueOccurMin()) {
                throw new UnexpectedValueException('The maximum number of values is strictly lower than the minimum');
            }
        }

        $this->valueOccurMax = $valueOccurMax;
        return $this;
    }
}