<?php
/**
 * Déclaration de la classe CommandLine\Argument\Parser\DecimalParser
 */
namespace jrosset\CommandLine\Argument\Parser;

use InvalidArgumentException;
use RangeException;

/**
 * Parseur vers un réel
 *
 * @package      CommandLine\Argument\Parser
 *
 * @noinspection PhpUnused
 */
class DecimalParser implements IParser {
    /**
     * @var double|null La valeur minimum autorisée. Null si pas de limite.
     */
    protected ?float $valueMin = null;
    /**
     * @var double|null La valeur maximum autorisée. Null si pas de limite
     */
    protected ?float $valueMax = null;

    /**
     * Crée un nouveau parseur
     *
     * @param double|null $valueMin La valeur minimum autorisée
     * @param double|null $valueMax La valeur maximum autorisée
     */
    public function __construct (?float $valueMin = null, ?float $valueMax = null) {
        $this->setValueMin($valueMin);
        $this->setValueMax($valueMax);
    }

    /**
     * @inheritDoc
     */
    public function parseValue (string $val) {
        if (!$this->isDecimal($val)) {
            throw new RangeException('La valeur n\'est pas un réel valide : ' . $val);
        }

        $int = (int)$val;

        if ($this->hasValueMin()) {
            if ($int < $this->getValueMin()) {
                throw new RangeException('La valeur est inférieure au minimum (' . $this->getValueMin() . ') : ' . $val);
            }
        }
        if ($this->hasValueMax()) {
            if ($int > $this->getValueMax()) {
                throw new RangeException($int, 'La valeur est supérieur au maximum (' . $this->getValueMax() . ') : ' . $val);
            }
        }

        return $int;
    }

    /**
     * Vérifie si la valeur est bien un nombre réel ?
     *
     * @return bool Est-ce que la valeur est bien un nombre réel ?
     */
    protected function isDecimal ($val): bool {
        if (!is_numeric($val)) {
            return false;
        }

        return true;
    }

    /**
     * @inheritDoc
     */
    public function getValueDescription (): string {
        return ($this->hasValueMin() ? $this->getValueMin() . ' <= ' : '') . 'decimal' . ($this->hasValueMax() ? ' <= ' . $this->getValueMax() : '');
    }

    /**
     * La valeur minimumm autorisée
     *
     * @return double|null La valeur minimum.
     *
     * @see $valueMin
     */
    public function getValueMin (): ?float {
        return $this->valueMin;
    }
    /**
     * Définit la valeur minimum autorisée.
     *
     * @param double|null $valueMin La valeur minimum.
     *
     * @return $this
     */
    public function setValueMin (?float $valueMin = null): self {
        if (!is_null($valueMin)) {
            if (!$this->isDecimal($valueMin)) {
                throw new InvalidArgumentException('Le minimum n\'est pas un entier ou null : ' . $valueMin);
            }
        }

        $this->valueMin = $valueMin;
        return $this;
    }
    /**
     * Est-ce qu'il existe une limite minimum ?
     *
     * @return boolean True s'il existe une limite, sinon False
     */
    public function hasValueMin (): bool {
        return !is_null($this->getValueMin());
    }

    /**
     * La valeur maximum autorisée
     *
     * @return double|null La valeur maximum.
     *
     * @see $valueMax
     */
    public function getValueMax (): ?float {
        return $this->valueMax;
    }
    /**
     * Définit la valeur maximum autorisée.
     *
     * @param double|null $valueMax La valeur maximum.
     *
     * @return $this
     */
    public function setValueMax (?float $valueMax = null): self {
        if (!is_null($valueMax)) {
            if (!$this->isDecimal($valueMax)) {
                throw new InvalidArgumentException('Le maximum n\'est pas un entier ou null : ' . $valueMax);
            }
        }

        $this->valueMax = $valueMax;
        return $this;
    }
    /**
     * Est-ce qu'il existe une limite maximum ?
     *
     * @return boolean True s'il existe une limite, sinon False
     */
    public function hasValueMax (): bool {
        return !is_null($this->getValueMax());
    }
}