<?php
/**
 * Déclaration de la classe CommandLine\Argument\Parser\PathParser
 */

namespace jrosset\CommandLine\Argument\Parser;

use InvalidArgumentException;
use UnexpectedValueException;

/**
 * Parseur d'un chemin (fichier ou répertoire)
 *
 * Renvoie un {@see stdStringClass} avec les attributs suivants :
 *  - *fullpath* : chemin complet
 *  - directory  : chemin du répertoire contenant
 *  - filename   : nom du fichier (extension inclue)
 *  - basename   : nom du fichier, sans l'extension
 *  - extension  : extension du fichier
 *
 * @package CommandLine\Argument\Parser
 */
class PathParser implements IParser {
    /**
     * @var string[] Les expressions régulières pour identifier un chemin
     */
    const REGEX_PATH = array(
        'unix'    => /** @lang PhpRegExp */
            '@^(?<directory>/*(?:[^/\n]+(?:/+[^/\n]+)*/+)?)(?<filename>(?<basename>[^/\n]+?)(?:\.(?<extension>[a-zA-Z0-9_]+)|/+)?)$@',
        'windows' => /** @lang PhpRegExp */
            '@^(?<directory>(?:(?:[a-zA-Z]:|[/])[/]|(?![\\\\/]))(?:[^<>:"/\\\\|?*]+(?:[\\\\/][^<>:"/\\\\|?*]+)*[\\\\/]+)?)(?<filename>(?<basename>[^<>:"/\\\\|?*]+?)(?:\.(?<extension>[a-zA-Z0-9_]+)|[/])?)$@',
    );

    /**
     * @var bool Est-ce que le fichier doit exister ?
     */
    protected bool $mustExist;

    /**
     * Crée un nouveau parseur
     *
     * @param bool $mustExists Est-ce que le chemin doit exister ?
     */
    public function __construct (bool $mustExists) {
        $this->setMustExists($mustExists);
    }

    /**
     * @inheritDoc
     */
    public function parseValue (string $val) {
        foreach (self::REGEX_PATH as $regex) {
            if (preg_match($regex, $val, $matches, PREG_UNMATCHED_AS_NULL)) {
                if ($this->mustExist && !file_exists($matches[0])) {
                    throw new UnexpectedValueException('Le chemin n\'existe pas : ' . $val);
                }

                $out = new stdStringClass('fullpath');
                $out->fullpath = $matches[0];
                $out->directory = $matches['directory'] ?? '';
                $out->filename = $matches['filename'] ?? '';
                $out->basename = $matches['basename'] ?? '';
                $out->extension = $matches['extension'] ?? '';

                return $out;
            }
        }

        throw new UnexpectedValueException('La valeur n\'est pas un chemin valide : ' . $val);
    }

    /**
     * @inheritDoc
     */
    public function getValueDescription (): string {
        return 'path';
    }

    /**
     * Est-ce que le chemin doit exister ?
     *
     * @return bool True si le chemin doit exister, sinon False
     *
     * @see          $_mustExists
     *
     * @noinspection PhpUnused
     */
    public function getMustExist (): bool {
        return $this->mustExist;
    }
    /**
     * Spécifie si le chemin doit exister ou non
     *
     * @param bool $mustExists True si le chemin doit exister, sinon False
     *
     * @return $this
     *
     * @see $_mustExists
     */
    public function setMustExists (bool $mustExists): self {
        if (!is_bool($mustExists)) {
            throw new InvalidArgumentException('Le contrôle si le chemin doit exister ou non n\'est pas booléen : ' . $mustExists);
        }

        $this->mustExist = $mustExists;
        return $this;
    }
}