<?php
/**
 * Déclaration de la classe CommandLine\Argument\Parser\RegexParser
 */
namespace jrosset\CommandLine\Argument\Parser;

use RangeException;

/**
 * Parseur vers une chaine de caractère correspondant à un motif
 *
 * Renvoie la liste des groupes capturants
 *
 * @package      CommandLine\Argument\Parser
 *
 * @noinspection PhpUnused
 */
class RegexParser implements IParser {
    /**
     * @var string La regex à respecter
     */
    protected string $regex = '';
    /**
     * @var string|null Le format affiché dans la description
     */
    protected ?string $format = '';

    /**
     * Crée un nouveau parseur
     *
     * @param string      $regex  Le motif à respecter
     * @param string|null $format Le format à afficher
     */
    public function __construct (string $regex, ?string $format = null) {
        $this->setRegex($regex);
        $this->setFormat($format);
    }

    /**
     * @inheritDoc
     */
    public function parseValue (string $val) {
        if (!preg_match($this->regex, $val, $matches)) {
            throw new RangeException('La valeur ne correspond pas au motif attendu : ' . $val);
        }

        return $matches;
    }

    /**
     * @inheritDoc
     */
    public function getValueDescription (): string {
        return is_null($this->getFormat()) ? $this->getRegex() : $this->getFormat();
    }

    /**
     * Le motif à respecter
     *
     * @return string Le motif
     *
     * @see $regex
     */
    public function getRegex (): string {
        return $this->regex;
    }
    /**
     * Modifie le motif à respecter
     *
     * @param string $regex Le nouveau motif
     *
     * @return $this
     */
    public function setRegex (string $regex): self {
        $this->regex = $regex;
        return $this;
    }

    /**
     * Le format à afficher
     *
     * @return string|null Le format
     *
     * @see $format
     */
    public function getFormat (): ?string {
        return $this->format;
    }
    /**
     * Modifie le format à afficher
     *
     * @param string|null $format Le nouveau format
     *
     * @return $this
     */
    public function setFormat (?string $format = null): self {
        $this->format = $format;
        return $this;
    }
}