<?php

namespace jrosset\CommandLine\Argument\Value;

use InvalidArgumentException;
use jrosset\CommandLine\Argument\AbstractArgument;
use UnexpectedValueException;

/**
 * Abstract base for value arguments
 */
abstract class AbstractValueArgument extends AbstractArgument implements IValueArgument {
    /**
     * @var int The minimum number of values
     */
    protected int $occurMin = 1;
    /**
     * @var int|null The maximum number of values. Null if unlimited
     */
    protected ?int $occurMax = 1;

    /**
     * Create a new value argument
     *
     * @param string      $name        The argument name in help
     * @param string|null $description The argument description in help
     * @param boolean     $optional    Optional argument ?
     * @param mixed       $default     The default value
     */
    protected function __construct (string $name, ?string $description, bool $optional = false, $default = null) {
        parent::__construct($name, $description);

        $this->setOptional($optional);
        $this->setDefault($default);
    }

    /**
     * Is the argument optional ?
     *
     * @return bool Is the argument optional ?
     */
    public function isOptional (): bool {
        return $this->getOccurMin() === 0;
    }
    /**
     * Set if the argument is optional
     *
     * @param bool $optional Is the argument optional ?
     *
     * @return $this
     */
    public function setOptional (bool $optional): self {
        $this->setOccurMin($optional ? 0 : 1);
        return $this;
    }

    /**
     * @inheritDoc
     */
    public function getOccurMin (): int {
        return $this->occurMin;
    }
    /**
     * Set the minimum number of occurrence of values
     *
     * @param int $occurMin The minimum number of occurrence of values
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the minimum number of occurrence of values provided is not valid (strictly positive integer)
     * @throws UnexpectedValueException If the minimum number of occurrence of values provided is strictly greater than the maximum
     */
    public function setOccurMin (int $occurMin): self {
        if (!is_numeric($occurMin) || floor($occurMin) != $occurMin || ($occurMin = (int)$occurMin) < 0) {
            throw new InvalidArgumentException('The minimum number of values must be a strictly positive integer');
        }

        $occurMax = $this->getOccurMax();
        if ($occurMax !== null && $occurMin > $occurMax) {
            throw new UnexpectedValueException('The minimum number of values provided is strictly greater than the maximum');
        }

        $this->occurMin = $occurMin;
        return $this;
    }

    /**
     * @inheritDoc
     */
    public function getOccurMax (): ?int {
        return $this->occurMax;
    }
    /**
     * Set the maximum number of occurrence of values. Null if unlimited
     *
     * @param int|null $occurMax The maximum number of occurrence of values
     *
     * @return $this
     *
     * @throws InvalidArgumentException If the maximum number of occurrence of values provided is not valid (strictly positive integer)
     * @throws UnexpectedValueException If the maximum number of occurrence of values provided is strictly lower than the minimum
     */
    public function setOccurMax (?int $occurMax): self {
        if ($occurMax !== null) {
            if (!is_numeric($occurMax) || floor($occurMax) != $occurMax || ($occurMax = (int)$occurMax) < 0) {
                throw new InvalidArgumentException('The maximum number of values must be a strictly positive integer');
            }

            if ($occurMax < $this->getOccurMin()) {
                throw new UnexpectedValueException('The maximum number of values is strictly lower than the minimum');
            }
        }

        $this->occurMax = $occurMax;
        return $this;
    }
}